#ifndef __CPersistPropertiesCaps__
#define __CPersistPropertiesCaps__

#include "IPersistData.h"
#include <qobject.h>
#include <qvariant.h>
#include <qmetaobject.h>

class CPersistPropertiesCaps : public IPersistData
{
private:
  QObject *m_Object;
  bool     m_bAll;
public:
  CPersistPropertiesCaps(QObject *q,bool all=false) : IPersistData()
    {
      m_Object=q;
      m_bAll=all;
    }

  void load(istream& s)
    {
      static char name[512];
      static char value[512];
      int         i,n,t;
      
      s >> n;
      for (i=0;i<n;i++)
      {
	stringFromStream(s,name);
	s >> t;
	if (stringFromStream(s,value) == 0)
	  continue;
	m_Object->setProperty(name,fromString(value,(QVariant::Type)t));
      }
    }
  
  void save(ostream& s)
    {
      QString     old,qstr;
      const char *str;

      QStrList lst=m_Object->metaObject()->propertyNames(m_bAll);
      for (int i=0;i<lst.count();i++)
      {
	// valid type ?
	if (!isValidType(m_Object->property((char *)lst.at(i)).type()))
	  lst.remove(i);
      }
      s << lst.count() << ' ';
      for (int i=0;i<lst.count();i++)
      {
	qstr=toString(m_Object->property((char *)lst.at(i)));
	stringToStream(s,QString(lst.at(i))); 
	s << ' '
	  << m_Object->property((char *)lst.at(i)).type()
	  << ' ';
	stringToStream(s,qstr);
	s << ' ';
      }
      s << endl;
    }
  
  void failed()
    {

    }

  // ------ methods from Xavier Decoret state holder
  
  static bool isValidType(const QVariant::Type& t)
    {
      switch (t)
      {
	// The following types are valid
      case QVariant::String:
      case QVariant::CString:    
      case QVariant::ByteArray:  
      case QVariant::Int:        
      case QVariant::UInt:       
      case QVariant::Bool:       
      case QVariant::Double:     
      case QVariant::Date:
      case QVariant::Time:
      case QVariant::DateTime:
      case QVariant::Rect:
	return true;
      }
      return false;
    }
  
  static QString toString(const QVariant& v)
    {
      switch (v.type())
      {
	// The following types knows how to convert to a string
      case QVariant::ByteArray:  
      case QVariant::Int:        
      case QVariant::UInt:       
      case QVariant::Bool:       
      case QVariant::Double:     
      case QVariant::Date:
      case QVariant::Time:
      case QVariant::DateTime:
      case QVariant::String:
      case QVariant::CString:    
	return v.toString();
	// The next types are the one for which we implement a converter
	// -> rect
      case QVariant::Rect:
	{
	  const QRect r=v.toRect();
	  return QString().sprintf("%dx%d%+d%+d",
				   r.width(),r.height(),r.left(),r.top());
	}
	// The next types are the non-handled ones
      case QVariant::Font:       
      case QVariant::Brush:      
      case QVariant::Color:      
      case QVariant::Palette:    
      case QVariant::Region:     
      case QVariant::Cursor:     
      case QVariant::KeySequence:
      case QVariant::BitArray:   
      case QVariant::ColorGroup: 
      case QVariant::Size:       
      case QVariant::StringList: 
      case QVariant::Pixmap:     
      case QVariant::Point:      
      case QVariant::Image:      
      case QVariant::PointArray: 
      case QVariant::Bitmap:     
      case QVariant::Map:        
      case QVariant::SizePolicy: 
      case QVariant::List:       
      case QVariant::IconSet:
      case QVariant::Invalid:
	break;
      }
      return ("");
    }

  static QVariant fromString(const char* s,const QVariant::Type& type)
    {
      switch (type)
      {
	// The following types knows how to convert from a QVariant(const char*)
      case QVariant::String:
      case QVariant::CString:    
      case QVariant::ByteArray:  
      case QVariant::Int:        
      case QVariant::UInt:       
      case QVariant::Double:     
      case QVariant::Date:
      case QVariant::Time:
      case QVariant::DateTime:
	{
	  QVariant v(s);
	  (void) v.cast(type);
	  return v;
	}
	// The next types are the one for which we implement a converter
      case QVariant::Bool:
	return QVariant(QVariant(s).toInt()).toBool();
      case QVariant::Rect:
	{
	  int left,top,width,height;
	  sscanf(s,"%dx%d%d%d",&width,&height,&left,&top);
	  return QVariant(QRect(left,top,width,height));
	}
	// The next types are the non-handled ones
      case QVariant::Font:       
      case QVariant::Brush:      
      case QVariant::Color:      
      case QVariant::Palette:    
      case QVariant::Region:     
      case QVariant::Cursor:     
      case QVariant::KeySequence:
      case QVariant::BitArray:   
      case QVariant::ColorGroup: 
      case QVariant::Size:       
      case QVariant::StringList: 
      case QVariant::Pixmap:     
      case QVariant::Point:      
      case QVariant::Image:      
      case QVariant::PointArray: 
      case QVariant::Bitmap:     
      case QVariant::Map:        
      case QVariant::SizePolicy: 
      case QVariant::List:       
      case QVariant::IconSet:
      case QVariant::Invalid:
	break;
      }
      return QVariant(QVariant::Invalid);
    }

  static void stringToStream(ostream& o,const QString& qstr)
    {
      const char *str;
      
      str=qstr.latin1();
      if (str == NULL)
	o << "0 ";
      else
	o << strlen(str) << ' ' << str;
    }

  static int stringFromStream(istream& i,char *str)
    {
      int         c,l,strp=0;
      
      // TODO: securiser debordement
      i >> l;
      if (l == 0)
	return (0);
      c=i.get();
      if (c != ' ')
	throw CException("CPersistPropertiesCaps: string syntax error !");
      for (;i && l>0;l--)
	str[strp++]=i.get();
      str[strp]='\0';	  
      if (!i)
	throw CException("CPersistPropertiesCaps: string syntax error !");
      return (strp);
    }
  
};


#endif
