// -------------------------------------------------------------------------
#ifdef WIN32
#include <windows.h>
#endif
#ifdef __APPLE__
	#include <OpenGL/gl.h>
#else
	#include <GL/gl.h>
#endif
#include "CVertex.h"
// -------------------------------------------------------------------------
#include <math.h>
#include <iostream>
// -------------------------------------------------------------------------
#ifndef ABS
  #define ABS(x) ((x)>0?(x):(-(x)))
#endif /* ABS */
// -------------------------------------------------------------------------
inline CVertex::CVertex()
{
  m_dX = m_dY = m_dZ = 0.0;
}
// -------------------------------------------------------------------------
inline CVertex::CVertex(const double	&x_,
		 const double	&y_,
		 const double	&z_)
{
  m_dX = x_;
  m_dY = y_;
  m_dZ = z_;
}
// -------------------------------------------------------------------------
inline CVertex::CVertex(const double	&x_,
		 const double	&y_)
{
  m_dX = x_;
  m_dY = y_;
  m_dZ = 0.0;
}
// -------------------------------------------------------------------------
inline double	CVertex::dot(const CVertex & a) const
{
  return (a.x() * m_dX + a.y() * m_dY + a.z() * m_dZ);
}
// -------------------------------------------------------------------------
inline CVertex	CVertex::mul(const CVertex & a) const
{
  return (CVertex(a.x() * m_dX,a.y() * m_dY,a.z() * m_dZ));
}
// -------------------------------------------------------------------------
inline CVertex	add(const CVertex &	a,const CVertex & b)
{
  return CVertex(a.x() + b.x(), a.y() + b.y(), a.z() + b.z());
}
// -------------------------------------------------------------------------
inline CVertex	sub(const CVertex &	a,
			  const CVertex &	b)
{
  return CVertex(a.x() - b.x(), a.y() - b.y(), a.z() - b.z());
}
// -------------------------------------------------------------------------
inline CVertex	neg(const CVertex & p)
{
  return CVertex(-p.x(), -p.y(), -p.z());
}
// -------------------------------------------------------------------------
inline CVertex	mul(const double	d,const CVertex	&p)
{
  return CVertex(d * p.x(), d * p.y(), d * p.z());
}
// -------------------------------------------------------------------------
inline CVertex	mul(const CVertex &p,const double	d)
{
  return (mul(d,p));
}
// -------------------------------------------------------------------------
inline CVertex	 mul(const double m[4][4], const CVertex & b)
{
  return (CVertex(m[0][0]*b.x()+m[1][0]*b.y()+m[2][0]*b.z()+m[3][0],
		  m[0][1]*b.x()+m[1][1]*b.y()+m[2][1]*b.z()+m[3][1],
		  m[0][2]*b.x()+m[1][2]*b.y()+m[2][2]*b.z()+m[3][2]));
}
// -------------------------------------------------------------------------
inline CVertex	 CVertex::normMult(const double m[4][4])
{
  return (CVertex(m[0][0]*x()+m[1][0]*y()+m[2][0]*z(),
		  m[0][1]*x()+m[1][1]*y()+m[2][1]*z(),
		  m[0][2]*x()+m[1][2]*y()+m[2][2]*z()));  
}
// -------------------------------------------------------------------------
inline CVertex	div(const CVertex &	p,double	d)
{
  return CVertex(p.x() / d, p.y() / d, p.z() / d);
}
// -------------------------------------------------------------------------
inline double	mul(const CVertex& a,const CVertex& b)
{
  return (a.dot(b));
}
// -------------------------------------------------------------------------
inline bool CVertex::equal(const CVertex& a) const
{
  return (ABS(x()-a.x()) < CVertex::getEpsilon() 
       && ABS(y()-a.y()) < CVertex::getEpsilon() 
       && ABS(z()-a.z()) < CVertex::getEpsilon());
}
// -------------------------------------------------------------------------
inline CVertex	CVertex::vect(const CVertex&	p) const
{
  return CVertex(m_dY * p.z() - p.y() * m_dZ, m_dZ * p.x() - p.z() * m_dX, m_dX * p.y() - p.x() * m_dY);
}
// -------------------------------------------------------------------------
inline CVertex	CVertex::cross(const CVertex &p) const
{
  return (vect(p));
}
// -------------------------------------------------------------------------
inline double CVertex::getEpsilon()
{
  return (s_dEpsilon);
}
// -------------------------------------------------------------------------
inline void CVertex::setEpsilon(double e)
{
  if (e > 0.0) 
    s_dEpsilon=e; 
  else 
    throw CException("CVertex: epsilon should be greater than 0.0 in CVertex::setEpsilon");
}
// -------------------------------------------------------------------------
inline void    CVertex::gl() const
{
  glVertex3d(m_dX,m_dY,m_dZ);
}
// -------------------------------------------------------------------------
inline double& CVertex::operator[](int i)
{
  switch (i)
  {
    case 0: return (m_dX);
    case 1: return (m_dY);
    case 2: return (m_dZ);
    default: throw CException("CVertex: index out of range in CVertex::[]");
  }
}
// -------------------------------------------------------------------------
inline const double& CVertex::operator[](int i) const
{
  switch (i)
  {
    case 0: return (m_dX);
    case 1: return (m_dY);
    case 2: return (m_dZ);
    default: throw CException("CVertex: index out of range in CVertex::[]");
  }
}
// -------------------------------------------------------------------------
inline void CVertex::axis(CVertex& u,CVertex& v) const
{
  u.set(-y(),x(),0.0); 
  if (u.length() < CVertex::getEpsilon())
  {
    u.set(z(),0.0,-x());
    if (u.length() < CVertex::getEpsilon())
      throw CException("CVertex: can't create axis");
  }
  u.unit();
  v=vect(u);
  v.unit();
}
// -------------------------------------------------------------------------
inline bool CVertex::operator== (const CVertex& v) const
{
	return (equal(v));
}
// -------------------------------------------------------------------------
inline bool CVertex::operator!= (const CVertex& v) const
{
	return (!equal(v));	
}
// -------------------------------------------------------------------------
inline CVertex& CVertex::operator+= (const CVertex& v)
{
	m_dX+=v.x();
	m_dY+=v.y();
	m_dZ+=v.z();
	return (*this);	
}
// -------------------------------------------------------------------------
inline CVertex& CVertex::operator-= (const CVertex& v)
{
	m_dX-=v.x();
	m_dY-=v.y();
	m_dZ-=v.z();
	return (*this);	
}
// -------------------------------------------------------------------------
inline CVertex& CVertex::operator*= (double v)
{
	m_dX*=v;
	m_dY*=v;
	m_dZ*=v;
	return (*this);	
}
// -------------------------------------------------------------------------
inline CVertex& CVertex::operator/= (double v)
{
	m_dX/=v;
	m_dY/=v;
	m_dZ/=v;
	return (*this);	
}
// -------------------------------------------------------------------------
inline CVertex CVertex::operator+ (const CVertex& v) const
{
	return CVertex(x()+v.x(),y()+v.y(),z()+v.z());
}
// -------------------------------------------------------------------------
inline CVertex CVertex::operator- (const CVertex& v) const
{
	return CVertex(x()-v.x(),y()-v.y(),z()-v.z());	
}
// -------------------------------------------------------------------------
inline CVertex CVertex::operator* (double v) const
{
	return CVertex(x()*v,y()*v,z()*v);	
}
// -------------------------------------------------------------------------
inline CVertex CVertex::operator/ (double v) const
{
	return CVertex(x()/v,y()/v,z()/v);
}
// -------------------------------------------------------------------------
inline CVertex CVertex::operator- () const
{
	return CVertex(-x(),-y(),-z());
}
// -------------------------------------------------------------------------
inline CVertex operator* (double s, const CVertex& v)
{
  return (v*s);
}
// -------------------------------------------------------------------------
inline double	CVertex::length() const
{
  return sqrt(m_dX*m_dX+m_dY*m_dY+m_dZ*m_dZ);
}
// -------------------------------------------------------------------------
inline double	CVertex::sqrdLength() const
{
  return (m_dX*m_dX+m_dY*m_dY+m_dZ*m_dZ);
}
// -------------------------------------------------------------------------
inline std::ostream& operator<<(std::ostream& s,const CVertex & v)
{
  return (s << "(" << v.x() << "," << v.y() << "," << v.z() << ")");
}
// -------------------------------------------------------------------------
inline std::istream& operator>>(std::istream& s,CVertex& v)
{
  int    i;
  double d[3];
  char   c;

  s >> c;
  if (c == '(')
  {
    for (i=0;i<3;i++)
    {
      s >> d[i] >> c;
      if (i < 2 && c != ',')
	s.clear(std::ios::badbit);
      else if (i == 2 && c != ')')
	s.clear(std::ios::badbit);
    }
  }
  else
  {
    s.clear(std::ios::badbit);
  }
  if (s)
    v.set(d[0],d[1],d[2]);
  else
    throw CException("CVertex::>> syntax error !");
  return (s);
}
// -------------------------------------------------------------------------
