// ----------------------------------------------------
#include "CViewer.h"
#include "CWorld.h"
#include "CInterface.h"
#include <math.h>


#ifdef QT3_SUPPORT
	#include <QCheckBox>
	#include <QButtonGroup>
	#include <QPushButton>
	#include <QStatusBar>
	#include <Q3Button>
	#include <QMouseEvent>
#else
	#include <qcheckbox.h>
	#include <qbuttongroup.h>
	#include <qpushbutton.h>
	#include <qstatusbar.h>
	#include <qbutton.h>
#endif
#include <CTexture.h>
// ----------------------------------------------------
#include "QGLViewer/camera.h"

#define ROT_CAMERA 0 // rotation camera incompatible avec ctrl angle
static int toggle_camera  = 0;
// ----------------------------------------------------
#include "CReceiver.h"

// retransmet au mobile camera les mouvements fait depuis l'interface
void Receiver::frameUpdated() {
    //float x,y,z;
	qglviewer::Vec v = g_MainWindow->m_ViewPort->camera()->position();
    WORLD.getCamera()->pos().setX(100*v.x);
    WORLD.getCamera()->pos().setY(100*v.y);
    WORLD.getCamera()->setHeight(2.5/v.z);
#if ROT_CAMERA
    float qx,qy,qz, sa,ca;
    g_MainWindow->m_ViewPort->camera()->getOrientation(qx,qy,qz,ca);
    sa = sqrt(qx*qx+qy*qy+qz*qz);
    if (sa) {
      WORLD.getCamera()->setVar(0,qx/sa);
      WORLD.getCamera()->setVar(1,qy/sa);
      WORLD.getCamera()->setVar(2,qz/sa);
    }
    WORLD.getCamera()->setAngle(2*atan2(sa,ca));
#endif
  }

// ----------------------------------------------------
void CViewer::resetDisplay() {
  camera()->setPosition(qglviewer::Vec(0,0,2.5));
  camera()->setOrientation(qglviewer::Quaternion(0,0,0,1));
  WORLD.getCamera()->setVisible(0);
  WORLD.getCamera()->setColor(0xffffff00);
  WORLD.getCamera()->setAngle(0.);
  WORLD.getCamera()->setVar(0,0.);
  WORLD.getCamera()->setVar(1,0.);
  WORLD.getCamera()->setVar(2,1.);
  toggle_camera = 0;
  WORLD.getLight1()->pos().setX(0);
  WORLD.getLight1()->pos().setY(0);
  WORLD.getLight1()->setHeight(1e8);
  WORLD.getLight1()->setVisible(1);
  WORLD.getLight1()->setColor(0xffffff00);
  WORLD.getLight2()->setVisible(1);
  WORLD.getLight2()->setColor(0x00000000);
  WORLD.getLight2()->pos().setX(0);
  WORLD.getLight2()->pos().setY(0);
  WORLD.getLight2()->setHeight(1e8);
  WORLD.getLight3()->setVisible(1);
  WORLD.getLight3()->setColor(0x00000000);
  WORLD.getLight3()->pos().setX(0);
  WORLD.getLight3()->pos().setY(0);
  WORLD.getLight3()->setHeight(1e8);
}

CViewer::CViewer(QWidget *parent,const char *name)
  : QGLViewer(parent, name)
{
  static Receiver *myReceiver = new Receiver();
  setMouseTracking(true);
  startAnimation();
  m_bPick=false;
  m_uiTex=0;
#if ROT_CAMERA
  setMouseStateKey(QGLViewer::CAMERA, Qt::ControlButton);
#endif
 
  resetDisplay();
  connect(camera()->frame(), SIGNAL(modified()), myReceiver, SLOT(frameUpdated()));
}
// ----------------------------------------------------
void CViewer::preDraw() {

  // --- effacer ou ne pas effacer l'ecran
  
  if (!WORLD.getCamera()->isVisible()) {
    glDrawBuffer(GL_BACK);
    setAutoBufferSwap(true);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);    
  }
  else {
    glDrawBuffer(GL_FRONT);
    setAutoBufferSwap(false);
    glFlush();
  }

  // --- eclairage

  unsigned int c; //  = WORLD.getCamera()->getColor();
  if (WORLD.getLight1()->isVisible() && (c=WORLD.getLight1()->getColor())) {
    glEnable(GL_LIGHT0);
    GLfloat col[4] = {
      ((c & 0xFF000000) >> 24)/255.0,
      ((c & 0x00FF0000) >> 16)/255.0,
      ((c & 0x0000FF00) >>  8)/255.0,
      1 };
    glLightfv(GL_LIGHT0,GL_DIFFUSE,col);
    GLfloat pos[4] = {
      WORLD.getLight1()->pos().x()/100.,
      WORLD.getLight1()->pos().y()/100.,
      WORLD.getLight1()->getHeight()/100.,
      1 };
    glLightfv(GL_LIGHT0,GL_POSITION, pos);
  }
  else  glDisable(GL_LIGHT1);
  
  if (WORLD.getLight2()->isVisible() && (c=WORLD.getLight2()->getColor())) {
    glEnable(GL_LIGHT1);
    GLfloat col[4] = {
      ((c & 0xFF000000) >> 24)/255.0,
      ((c & 0x00FF0000) >> 16)/255.0,
      ((c & 0x0000FF00) >>  8)/255.0,
      1 };
    glLightfv(GL_LIGHT1,GL_DIFFUSE,col);
    GLfloat pos[4] = {
      WORLD.getLight2()->pos().x()/100.,
      WORLD.getLight2()->pos().y()/100.,
      WORLD.getLight2()->getHeight()/100.,
      1 };
    glLightfv(GL_LIGHT1,GL_POSITION, pos);
  }
  else glDisable(GL_LIGHT1);
  
  if (WORLD.getLight3()->isVisible() && (c=WORLD.getLight3()->getColor())) {
    glEnable(GL_LIGHT2);
    GLfloat col[4] = {
      ((c & 0xFF000000) >> 24)/255.0,
      ((c & 0x00FF0000) >> 16)/255.0,
      ((c & 0x0000FF00) >>  8)/255.0,
      1 };
    glLightfv(GL_LIGHT2,GL_DIFFUSE,col);
    GLfloat pos[4] = {
      WORLD.getLight3()->pos().x()/100.,
      WORLD.getLight3()->pos().y()/100.,
      WORLD.getLight3()->getHeight()/100.,
      1 };
    glLightfv(GL_LIGHT2,GL_POSITION, pos);
  }
  else glDisable(GL_LIGHT2);
  
  // --- position et orientation de la camera

  camera()->setPosition(qglviewer::Vec(WORLD.getCamera()->pos().x()/100.,
			WORLD.getCamera()->pos().y()/100.,
			2.5/WORLD.getCamera()->getHeight()));
  double a = WORLD.getCamera()->getAngle();
  double sa = sin(a/2), ca = cos(a/2);
#if 1 // ROT_CAMERA
  double ax = WORLD.getCamera()->getVar(0),
         ay = WORLD.getCamera()->getVar(1),
         az = WORLD.getCamera()->getVar(2);
  camera()->setOrientation(qglviewer::Quaternion(sa*ax,sa*ay,sa*az, ca));
#else
  camera()->setOrientation(0,0,sa,ca);
#endif
  // GL_PROJECTION matrix
  camera()->loadProjectionMatrix();
  // GL_MODELVIEW matrix
  camera()->loadModelViewMatrix();
}
// ----------------------------------------------------
void CViewer::draw()
{
  // --- recuperation des coords de la souris dans le repere monde
  if (m_uiTex == 0)
  {
	glEnable(GL_TEXTURE_2D);
    static GLubyte raw[2*2*4]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    glGenTextures(1,&m_uiTex);
    glBindTexture(GL_TEXTURE_2D,m_uiTex);
    glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,2,2,0,GL_RGBA,GL_UNSIGNED_BYTE,raw);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glDisable(GL_TEXTURE_2D);
  }
  if (m_bPick)
  {
    double x,y;

    if (getTexCoordsAt(m_dXPick,m_dYPick,x,y))
    {
      WORLD.getMouse()->pos().setX(x*WORLD_SCALE);
      WORLD.getMouse()->pos().setY(-y*WORLD_SCALE);
      // Q: dans l'ideal, il faudrait tenir compte de la perspective.      $FAB
      // R: on en tient deja compte car les coords de texture sont dans le plan $SYLVAIN
      WORLD.getMouse()->speed().setX((x-m_dXMouse)*WORLD_SCALE);
      WORLD.getMouse()->speed().setY(-(y-m_dYMouse)*WORLD_SCALE); // $FAB 13/12/02
      m_dXMouse=x;
      m_dYMouse=y;
    }
    m_bPick=false;
  }
  else 
  { // $FAB 13/12/02
    // pb si artan2(dy,dx) !
    //WORLD.getMouse()->speed().setX(0.0);	
    //WORLD.getMouse()->speed().setY(0.0);
  }

  // --- reinitialisations a la nouvelle trame 
  WORLD.resetKeys();
  WORLD.sampleTime();
  
  // --- trace'
  glPushAttrib(GL_ENABLE_BIT);

  glEnable(GL_LIGHTING);
  glEnable(GL_NORMALIZE);
  // glEnable(GL_RESCALE_NORMAL);
  glEnable(GL_COLOR_MATERIAL);

  // draw ground
  glColor3d(1.0,1.0,1.0);
  glBegin(GL_QUADS);
  glNormal3d(0.0,0.0,1.0);
  glVertex3d(-1.0,-1.0,0.0);
  glVertex3d(-1.0,1.0,0.0);
  glVertex3d(1.0,1.0,0.0);
  glVertex3d(1.0,-1.0,0.0);
  glEnd();

  WORLD.drawMobiles();

  if (g_MainWindow->m_cbNames->isChecked())
  {
    // draw frame
    WORLD.drawFrame();
    // draw names
    WORLD.drawNames();
  }
  if (WORLD.drawState())
	  g_MainWindow->statusBar()->message("Disconnected by Server");

  glPopAttrib();

  g_MainWindow->refreshDisplay();
}
// ----------------------------------------------------
void CViewer::animate()
{
  NETWORK.processMessages();
  WORLD.step();
}
// ----------------------------------------------------
void CViewer::mouseMoveEvent(QMouseEvent *e)
{
  //if (control)
    QGLViewer::mouseMoveEvent(e);
    //else {
    m_dXPick=e->x();
    m_dYPick=e->y();
    m_bPick=true;
    //}
}
// ---------------------------------------------------------
void CViewer::mousePressEvent(QMouseEvent *e)
{

#if ROT_CAMERA
  if ((e->button() != Qt::LeftButton)
	  || (e->state()&Qt::ControlButton) )
#else
	if (e->button() != Qt::LeftButton)
#endif
      QGLViewer::mousePressEvent(e);
    else {
      WORLD.setClic(1);
      WORLD.setClicked(1);
    }
}
// ---------------------------------------------------------
void CViewer::mouseReleaseEvent(QMouseEvent *e)
{
#if ROT_CAMERA
  if ((e->button() != Qt::LeftButton)
      || (e->state()&Qt::ControlButton) )
#else
	if (e->button() != Qt::LeftButton)
#endif
    QGLViewer::mouseReleaseEvent(e);
  else {
    WORLD.setUnClic(1);
    WORLD.setClicked(0);
  }
}
// ---------------------------------------------------------
void CViewer::keyPressEvent(QKeyEvent *e)
{
  if (e->state()==Qt::ControlButton)
    switch (e->key())
      {
      case Qt::Key_L: resetDisplay();
	break;
      case Qt::Key_U:
	 if (g_MainWindow->hotkeys())
	   g_MainWindow->toggleSave();
	 break;
      case Qt::Key_P:
	 if (g_MainWindow->hotkeys())
	   g_MainWindow->togglePresets();
	 break;
      case Qt::Key_3:
      case Qt::Key_C:
	toggle_camera=1-toggle_camera;
	if (toggle_camera) {
	  WORLD.getCamera()->setVar(0,1.);
	  WORLD.getCamera()->setVar(1,0.);
	  WORLD.getCamera()->setVar(2,0.);
	  WORLD.getCamera()->setHeight(1.4);    // 1.5
	  WORLD.getCamera()->pos().setY(-150.); // -220
	  WORLD.getCamera()->setAngle(.6);      // .9
	}
	else {
	  WORLD.getCamera()->setVar(0,0.);
	  WORLD.getCamera()->setVar(1,0.);
	  WORLD.getCamera()->setVar(2,1.);
	  WORLD.getCamera()->setHeight(1.);
	  WORLD.getCamera()->pos().setY(0.);
	  WORLD.getCamera()->setAngle(0.);
	}
	break;
    default:
      QGLViewer::keyPressEvent(e);
      }
  else
    switch (e->key())
    {
    case Qt::Key_Left : WORLD.setKey(KEY_LEFT);  break;
    case Qt::Key_Right: WORLD.setKey(KEY_RIGHT); break;
    case Qt::Key_Up   : WORLD.setKey(KEY_UP);    break;
    case Qt::Key_Down : WORLD.setKey(KEY_DOWN);  break;
    case Qt::Key_Space: WORLD.setKey(KEY_SPACE); break;
    default:
      QGLViewer::keyPressEvent(e);
    }
}
void CViewer::keyReleaseEvent(QKeyEvent *e)
{
  switch (e->key())
    {
    default:
      QGLViewer::keyReleaseEvent(e);
    }
}

// ---------------------------------------------------------
/**
 * 
 * @param x 
 * @param y 
 * @param u 
 * @param v 
 * @return 
 */
bool CViewer::getTexCoordsAt(double x,double y,double& u,double& v)
{
  int          nb;
  GLint        vp[4];
  GLfloat      feedback[512];
  bool         found=false;

  glPushAttrib(GL_ENABLE_BIT);
  // check tex coords of dr atlast selection pos on screen
  // use feedback buffer
  glFeedbackBuffer(512,GL_3D_COLOR_TEXTURE,feedback);
  glRenderMode(GL_FEEDBACK);
  // projection
  // load view transformation
  // pick matrix
  glGetIntegerv(GL_VIEWPORT,vp);
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  gluPickMatrix(x,vp[3]-y,4.0,4.0,vp);  // $FAB  13/12/02 : vp[3]-
  camera()->loadProjectionMatrix(false);

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  camera()->loadModelViewMatrix();

  // draw textured quad 	// $Michaël : bug under linux
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D,m_uiTex);
  glBegin(GL_QUADS);
  glTexCoord2d(-1.0,-1.0);
  glVertex3d(-1.0,-1.0,0.0);
  glTexCoord2d(-1.0,1.0);
  glVertex3d(-1.0,1.0,0.0);
  glTexCoord2d(1.0,1.0);
  glVertex3d(1.0,1.0,0.0);
  glTexCoord2d(1.0,-1.0);
  glVertex3d(1.0,-1.0,0.0);
  glEnd();

  // termine
  glFlush();
  // parse feedback
  nb=glRenderMode(GL_RENDER);
  if (nb)
  {
    int j,n,pos=0,vsize=11;
    while (pos < nb && !found)
    {
      switch ((int)feedback[pos++])
      {
      case GL_LINE_TOKEN:
      case GL_LINE_RESET_TOKEN:
	pos+=vsize*2;
	break;
      case GL_POINT_TOKEN:
      case GL_DRAW_PIXEL_TOKEN:
      case GL_COPY_PIXEL_TOKEN:
      case GL_BITMAP_TOKEN:
	pos+=vsize;
	break;
      case GL_PASS_THROUGH_TOKEN:
	pos++;
	break;
      case GL_POLYGON_TOKEN:
	n=(int)feedback[pos];
	for (j=0;j<n;j++)
	{
	  u=feedback[pos+1+vsize*j+7];
	  v=feedback[pos+1+vsize*j+8]; 
	  v=-v; // $FAB 13/12/02
	  found=true;
	  break;
	}
	pos+=1+n*vsize;
	break;
      }
    }
  }

  glPopAttrib();

  glMatrixMode(GL_PROJECTION);
  glPopMatrix();

  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();

  return (found);
}
// ----------------------------------------------------
