#include "UDistance.h"
using namespace UDistance;

#ifdef WIN32
// don't ask why, Visual .NET STL wants a comparison operator
// for pairs of any type (header "utility")
bool operator < (const CVertex& v0,const CVertex& v1) {return false;}
#endif

// ----------------------------------
UDistance::segment::segment(point _p0,point _p1)
{
  p0=_p0;
  p1=_p1;
  d=(p1-p0);
  if (p1 != p0)
    t=d.normalize();
  else
    t=0.0;
}

UDistance::line::line(point _p,point _d)
{
  p=_p;
  d=_d;
  d.normalize();
}
    
UDistance::line::line (const segment& s)
{
  p=s.p0;
  d=s.d;
}
// ----------------------------------
const UDistance::hit& UDistance::_min(const hit& h0,const hit& h1)
{
  if ((h0.first) < (h1.first))
    return (h0);
  else
    return (h1);
} 

// ----------------------------------
// Distance functions

hit UDistance::dist2dsqr(const point& p0,const point& p1)
{
  CVertex u=p1-p0;
  double l2=u.dot(u);
  if (l2 < CVertex::getEpsilon2())
    return (hit(l2,CVertex(0.0,0.0,0.0)));
  else
    return (hit(l2,-u.unit()));      
}

hit UDistance::dist2dsqr(const point& p,const line& l)
{
  double dot=(p-l.p).dot(l.d);
  point  pl=(l.p+l.d*dot);
  pl=pl-p;
  double l2=pl.dot(pl);
  if (l2 < CVertex::getEpsilon2())
    return (hit(l2,CVertex(0.0,0.0,0.0)));
  else
    return (hit(l2,-pl.unit()));      
}

hit UDistance::dist2dsqr(const point& p,const segment& s)
{
  hit d2s0,d2s1;
  
  double dot0,dot1;
  dot0=(p-s.p0).dot(s.d);
  dot1=(p-s.p1).dot(s.d);
  if (dot0*dot1 <= 0.0)
    {
      line   l(s);
      return (dist2dsqr(p,l));
    }
  else
    {
      d2s0=dist2dsqr(p,s.p0);
      d2s1=dist2dsqr(p,s.p1);
      return (_min(d2s0,d2s1));
    }
}

hit UDistance::dist2dsqr(const segment& s0,const segment& s1)
{
  hit d2s0s1p0,d2s0s1p1,d2s1s0p0,d2s1s0p1;
  
  // test for intersection
  CVertex cr00=(s0.p0-s1.p0).cross(s1.d);
  CVertex cr01=(s0.p1-s1.p0).cross(s1.d);
  CVertex cr10=(s1.p0-s0.p0).cross(s0.d);
  CVertex cr11=(s1.p1-s0.p0).cross(s0.d);
  if (cr00.z() * cr01.z() < 0.0 && cr10.z() * cr11.z() < 0.0)
    return (hit(0.0,CVertex(0.0,0.0,0.0)));
  else
    {
      // find closest pair
      d2s0s1p0=dist2dsqr(s1.p0,s0);
      d2s0s1p0.second=-d2s0s1p0.second;
      d2s0s1p1=dist2dsqr(s1.p1,s0);
      d2s0s1p1.second=-d2s0s1p1.second;
      d2s1s0p0=dist2dsqr(s0.p0,s1);
      d2s1s0p1=dist2dsqr(s0.p1,s1);
      return (_min(_min(d2s0s1p0,d2s0s1p1),_min(d2s1s0p0,d2s1s0p1)));
    }
}

// ----------------------------------------------------
