header "pre_include_hpp" 
{
#include <cmath>
#include <map>
#include <string>
class CMobile;
#include "atr_nfo.h"
}

header "pre_include_cpp" 
{
#include "CException.h"
#include "MyASTNode.h" // $FAB
#include "CMobile.h"
#include "CWorld.h"
#include "messages_calc.h"

extern int cur_line;
extern int cur_mob;
#define getCurLine() (cur_line+1) 
#define warning(msg) display_warning(msg,m_Mobile_id+1,getCurLine())
    
#define ME            	        -1
#define MOUSE          		-2
#define COLLIDER       		-3
#define PREV           		-4
#define NEXT       	    	-5
#define NEAREST        		-6
#define NEAREST2       		-7
#define NEAREST3       		-8
#define NEAREST_LEFT   		-9
#define NEAREST_RIGHT 	       -10
#define NEAREST_UP    	       -11
#define NEAREST_DOWN  	       -12
#define EMPTY         	       -13
#define NEAREST_INDIR 	       -14
#define NEAREST_INDIR2 	       -15
#define NEAREST_INDIR3	       -16
#define CAMERA                 -17
#define LIGHT1                 -18
#define LIGHT2                 -19
#define LIGHT3                 -20

// flags pour get*Ptr()
#define MOVEABLE               0x01
#define NOT_PSEUDOMOBILES      0x02
#define NOT_ME                 0x04
#define NOT_ME_IF_START      ( (field==FIELD_START)?NOT_ME:0 )

// flags pour get*Ref()
#define CREATE_ON_FAIL 0x01
#define MUST_EXIST     0x02

// macros utiles
#define is_local(ib) ( ((ib) == -1) || ((ib)-1 == m_Mobile_client_id) )

#define is_me(ia,ib) ( is_local(ib) &&      \
          ( ( (ia) == -1) || ( (ia)-1 == m_Mobile_id ) )    )

#define conv_ref_to_number(mob) (                       \
   ( is_local((mob).client) && ((mob).mobile>-1) )      \
        ? (mob).mobile                                  \
        : WORLD.getMobileLocalId(getMobilePtr(mob))  )


#define RED     "4278190080" // 0xff000000
#define GREEN     "16711680" // 0x00ff0000
#define BLUE         "65280" // 0x0000ff00
#define YELLOW  "4294901760" // 0xffff0000
#define CYAN      "16776960" // 0x00ffff00
#define PURPLE  "4278255360" // 0xff00ff00
#define WHITE   "4294967040" // 0xffffff00
#define BLACK            "0" // 0x00000000
#define GRAY    "2155905024" // 0x80808000
#define ORANGE  "4287102976" // 0xff880000
}


options 
{
  language="Cpp";
}

/*  **************************************************************
    LE PARSEUR D'AST (execute les noeuds une fois analyse's)
*/

{
  // --------------- routines utilitaires ---------------------
  
  // --- pour "attrb ref : value"   (lvalue)
  // identification (ds nfo) de l'attrb du mobile ref = (i,j)
    void CalcWalker::set_atr_nfo(atr_nfo &nfo, int i, int j, string nom, int var)
    {
        CMobile *m=NULL;

        int ia = i; // mob.mobile;
        int ib = j; // mob.client;
        bool pseudo = false;
	
        if( ia < 0 )
        {
            switch(ia)
            {
	    case ME:              // moi
	        m = m_Mobile; break;
            case CAMERA :         // camera
                m = WORLD.getCamera(); pseudo = true; break;
            case LIGHT1 :         // lampe 1
                m = WORLD.getLight1(); pseudo = true; break;
            case LIGHT2 :         // lampe 2
                m = WORLD.getLight2(); pseudo = true; break;
            case LIGHT3 :         // lampe 3
                m = WORLD.getLight3(); pseudo = true; break;
            case PREV :          // get previous mobile
                m=WORLD.getPrevMobile(m_Mobile);  break;
            case NEXT :          // get next mobile
                m=WORLD.getNextMobile(m_Mobile);  break;
	    case EMPTY:          // vide
	        m = NULL; break;
            default: break;
            }
        } 
        else if (ib > -1)
        {   // mobile on distant client
            throw CException(MSG_modif_another);
        }
        else
        { // mobile on local client
	  if ( ia == 0 )
	    {
	      m=NULL;
	      warning(MSG_null_ref);
	    }
	  else if (ia > -1)
	      m=WORLD.getLocalMobile(ia-1);
        }

	nfo.m = m;
	nfo.pseudo = pseudo;
        nfo.name = nom;
        nfo.variable = var;
    }

    // --- affectation des attributs (lvalue)
    void CalcWalker::setAttributes(atr_nfo& atr, double a)
    {
      if (atr.name == "x") {
	atr.m->pos().setX(a);
      }
      else if (atr.name == "y") {
	atr.m->pos().setY(a);
      }
      else if (atr.name == "dx") {
	atr.m->speed().setX(a);
      }
      else if (atr.name == "dy") {
	atr.m->speed().setY(a);
      }
      else if (atr.name == "angle") {
	atr.m->setAngle(a);
      }
      else if (atr.name == "icon") {
	atr.m->setIcon((int) a);
      }
      else if (atr.name == "mem") {
	atr.m->setVar(atr.variable, a);
      }
      else if (atr.name == "color") {
	atr.m->setColor((unsigned int)a);
      }
      else if (atr.name == "red") {
	unsigned int c;
	c = (unsigned int)( (a<=0)?0: (a>=1.)? 255.: 255.*a);
	atr.m->setColor( (atr.m->getColor() & 0x00ffff00) | ( c << 24 ) );
      }
      else if (atr.name == "green") {
	unsigned int c;
	c = (unsigned int)( (a<=0)?0: (a>=1.)? 255.: 255.*a);
	atr.m->setColor( (atr.m->getColor() & 0xff00ff00) | ( c << 16 ) );
      }
      else if (atr.name == "blue") {
	unsigned int c;
	c = (unsigned int)( (a<=0)?0: (a>=1.)? 255.: 255.*a);
	atr.m->setColor( (atr.m->getColor() & 0xffff0000) | ( c <<  8 ) );
      }
      else if (atr.name == "gray") {
	unsigned int c;
	c = (unsigned int)( (a<=0)?0: (a>=1.)? 255.: 255.*a);
	atr.m->setColor( (atr.m->getColor() & 0x00ffff00) | ( c << 24 ) );
	atr.m->setColor( (atr.m->getColor() & 0xff00ff00) | ( c << 16 ) );
	atr.m->setColor( (atr.m->getColor() & 0xffff0000) | ( c <<  8 ) );
      }
      else if (atr.name == "width") {
	atr.m->setWidth(a);
      }
      else if (atr.name == "height") {
	atr.m->setHeight(a);
      }
      else if (atr.name == "zoom") {
	atr.m->setWidth(a);
	atr.m->setHeight(a);
      }
      else if (atr.name == "visible") {
	atr.m->setVisible(a != 0.0);
	if (!atr.pseudo) atr.m->updateVisible();
      }
      else if (atr.name == "contact") {
	// nothing to do.
      }
      else
	throw CException(MSG_undef_attrb);
    }

    // --- retourne un pointeur sur un mobile (i, i@j, ou special)
    CMobile *CalcWalker::getMobilePtr(mob_adress mob, int mode)
    {
        CMobile *m=NULL;

        int ia = mob.mobile;
        int ib = mob.client;

        if( ia == -100)
            return NULL;

        if( ia < 0 )
        {
	  switch(ia)
            {
	   case ME:              // moi
	        if (mode&NOT_ME) throw CException(MSG_notme);
	        m = m_Mobile;          break;
           case MOUSE :          // mouse
	        if (mode&NOT_PSEUDOMOBILES) throw CException(MSG_notmouse);
	        if (mode&MOVEABLE)   throw CException(MSG_notmouse);
	        //if (mode&MOVEABLE)  { cerr<<cur_line; throw MyException(MSG_tst_notmouse);}
                m = WORLD.getMouse();  break;
           case CAMERA :         // camera
	        if (mode&NOT_PSEUDOMOBILES) throw CException(MSG_notcamera);
                m = WORLD.getCamera(); break;
           case LIGHT1 :         // lampe 1
	        if (mode&NOT_PSEUDOMOBILES) throw CException(MSG_notlight);
                m = WORLD.getLight1(); break;
           case LIGHT2 :         // lampe 2
	        if (mode&NOT_PSEUDOMOBILES) throw CException(MSG_notlight);
                m = WORLD.getLight2(); break;
           case LIGHT3 :         // lampe 3
	        if (mode&NOT_PSEUDOMOBILES) throw CException(MSG_notlight);
                m = WORLD.getLight3(); break;
            case COLLIDER :      // collider
                m = m_Collider;        break;
            case PREV :          // get previous mobile
                m=WORLD.getPrevMobile(m_Mobile);  break;
            case NEXT :          // get next mobile
                m=WORLD.getNextMobile(m_Mobile);  break;
            case NEAREST :       // plusproche
                m = WORLD.getNearest(m_Mobile);   break;
            case NEAREST2 :      // plusproche
                m = WORLD.getNearest(m_Mobile,2); break;
            case NEAREST3 :      // plusproche
                m = WORLD.getNearest(m_Mobile,3); break;
            case NEAREST_LEFT :  // plusproche_gauche
                m = WORLD.getNearestDir(m_Mobile,-1,0,M_PI/2); break;
            case NEAREST_RIGHT : // plusproche_droit
                m = WORLD.getNearestDir(m_Mobile,1,0,M_PI/2);  break;
            case NEAREST_UP :    // plusproche_haut
                m = WORLD.getNearestDir(m_Mobile,0,1,M_PI/2);  break;
            case NEAREST_DOWN :  // plusproche_bas
                m = WORLD.getNearestDir(m_Mobile,0,-1,M_PI/2); break;
	    case EMPTY:          // NULL
	        m = NULL;            break;
	    case NEAREST_INDIR:
		m = WORLD.getNearestDir(m_Mobile, mob.dx, mob.dy, mob.angle);
		break;
	    case NEAREST_INDIR2:
		m = WORLD.getNearestDir(m_Mobile, mob.dx, mob.dy, mob.angle, 2);
		break;
	    case NEAREST_INDIR3:
		m = WORLD.getNearestDir(m_Mobile, mob.dx, mob.dy, mob.angle, 3);
		break;
            default: break;
            }
        } 
        else if (ib > -1)
        {   // mobile on distant client
            if (ia == -1) throw CException(MSG_mobile_id);
	    if ((mode&NOT_ME) && is_me(ia,ib)) throw CException(MSG_notme);
            m = WORLD.getDistantMobile(ia-1,ib-1);
        }
        else
        {   // mobile on local client
	    if (ia == 0)
	        m = NULL;
	    else if (ia > -1)
            {
	        if ((mode&NOT_ME) && is_me(ia,-1))
		  throw CException(MSG_notme);
                m = WORLD.getLocalMobile(ia-1);
            }
        }
        
        return m;
    }

    // --- teste si deux ref sont egales.
    // NB: comme les INT sont pris pour des refs, cette fonction
    // permet au passage de traiter correctement leur comparaison.
    bool CalcWalker::areMobilesEqual(mob_adress mob, mob_adress mob2)
    {
      // std::cerr << "[DEBUG] in are_M_equal\n";
      if (  (mob.mobile>=0) && is_local(mob.client)   &&
           (mob2.mobile>=0) && is_local(mob2.client) )
	return (mob.mobile == mob2.mobile);
      else
	return (getMobilePtr(mob) == getMobilePtr(mob2));
    }

// --- teste ref=fexp (cast)
    bool CalcWalker::isMobileEqualN(mob_adress mob, double f)
      { int i = (int)f;
	// std::cerr << "[DEBUG] in is_M_eql_N" << f << i << (f==i) <<"\n";
	if (f!=i) return false;  // f n'est pas un int. (compare int et double !)
	if ( ( mob.mobile>=0) && is_local(mob.client) )
	  return (mob.mobile == i);
	else {
	  mob_adress mob2;
	  mob2.mobile=i; mob2.client=-1;
	  return (getMobilePtr(mob)==getMobilePtr(mob2));
	  // return (getMobilePtr(mob)==WORLD.getLocalMobile(i-1));
	}
      }

// --- retourne la valeur de l'attribut attrib du mobile mob (rvalue)
    double CalcWalker::get_atr_nfo(int attrib, mob_adress mob, int var)
    {
        double r;
        CMobile *m= getMobilePtr(mob);
        // give its value to the attribute
        r = 0.0;
	if (!m) warning(MSG_null_ref);
        else 
        {
	  switch( attrib )
            {
            case ATTR_X:  r=m->pos().x();   break;
            case ATTR_Y:  r=m->pos().y();   break;
            case ATTR_DX: r=m->speed().x(); break;
            case ATTR_DY: r=m->speed().y(); break;
            case ATTR_ANGLE:   r=m->getAngle();  break;
            case ATTR_ICONE:   r=m->getIcon();   break;
            case ATTR_VAR:     r=m->getVar(var); break;
            case ATTR_COULEUR: r=m->getColor();  break;
            case ATTR_ROUGE: r= ( ( m->getColor() & 0xff000000 ) >> 24 )/255.; break;
            case ATTR_VERT:  r= ( ( m->getColor() & 0x00ff0000 ) >> 16 )/255.; break;
            case ATTR_BLEU:  r= ( ( m->getColor() & 0x0000ff00 ) >>  8 )/255.; break;
            case ATTR_GRIS:  {
				r  =  ( ( m->getColor() & 0x0000ff00 ) >>  8);
				r +=  ( ( m->getColor() & 0x00ff0000 ) >> 16);
				r +=  ( ( m->getColor() & 0xff000000 ) >> 24);
				r /= (3.0*255.);
				break;
			     }
            case ATTR_LARGEUR: r=m->getWidth();  break;
            case ATTR_HAUTEUR: r=m->getHeight(); break;
            case ATTR_CHOCX:   r=m_Choc.x();     break;
            case ATTR_CHOCY:   r=m_Choc.y();     break;
            case ATTR_VISIBLE: r= ( m->isVisible()    ? 1.0 : 0.0 ); break;
            case ATTR_CONTACT: r= ( m->firstContact() ? 1.0 : 0.0 ); break;

            default: r = 0.0; break;
            }
        }
        return r;
    }

    // retourne un pointeur sur une variable temporaire float
    double *CalcWalker::getFloatRef(string s, int mode)
    {
        fltMap_t::iterator it = mFloats.find(s);
        if( it != mFloats.end() )
        {
            return &((*it).second);
        }
        else if( mode & CREATE_ON_FAIL )
        {
            mFloats[s]= 0.0;
            return &(mFloats[s]);
        }
        else {
	    if (mode & MUST_EXIST ) throw CException(MSG_undef_f);
            return NULL;
	}
    }
    
    // retourne un pointeur sur une variable temporaire ref
    mob_adress *CalcWalker::getMobileRef(string s, int mode)
    {
        mobMap_t::iterator it = mMobiles.find(s);
        if( it != mMobiles.end() )
        {
            return &((*it).second);
        }
        else if( mode & CREATE_ON_FAIL )
        {
            mob_adress a;
            a.mobile = -1; a.client = -1;
            mMobiles[s]=a;
            return  &(mMobiles[s]);
        }
        else {
	    if (mode & MUST_EXIST) throw CException(MSG_undef_r);
            return NULL;
	}
    }

// versions secure pour l'ecriture des variables:
// detruisent d'abord la variable si elle existe dans l'autre type
    double *CalcWalker::getFloatWriteRefCheck(string s, int mode) {
      if (getMobileRef(s)) {
	mMobiles.erase(s);
	// cerr << "[DEBUG] remove rvar\n";
      }
      return getFloatRef(s, mode);
    }
    mob_adress *CalcWalker::getMobileWriteRefCheck(string s, int mode) {
      if (getFloatRef(s)) {
	mFloats.erase(s);
	// cerr << "[DEBUG] remove fvar\n";
      }
      return getMobileRef(s, mode);
    }

}




class CalcWalker extends TreeParser;
// options {defaultErrorHandler=false;}
{
private:
    int field;
    CMobile *m_Mobile, *m_Collider;
    int m_Mobile_id,m_Mobile_client_id;
    CVertex m_Choc;
    typedef std::map<string, double> fltMap_t;
    fltMap_t mFloats;
    typedef std::map<string, mob_adress> mobMap_t;
    mobMap_t mMobiles;

    class snapshot
    {
        public:
        fltMap_t mFloats;
        mobMap_t mMobiles;
        mobile_data mMobileData;
        public:
        snapshot() : mFloats(), mMobiles() {}
    };
    list<snapshot> mMemory;

public:
    void bindTo(CMobile *m, int fld, CMobile *collider=NULL, CVertex choc = CVertex(0.0, 0.0, 0.0))
    {
      m_Mobile=m; 
      m_Mobile_id = WORLD.getMobileLocalId(m)-1;
      m_Mobile_client_id = NETWORK.clientId();
      cur_line = 0;
      cur_mob = m_Mobile_id+1;
      field = fld;
      m_Collider = collider;
      mFloats.clear();
      mMobiles.clear();
      mMemory.clear();
      m_Choc = choc;
    }
    double get_atr_nfo(int, mob_adress, int);
    void set_atr_nfo(atr_nfo&, int, int, string, int);
    CMobile *getMobilePtr(mob_adress, int mode=0);
    bool areMobilesEqual(mob_adress mob, mob_adress mob2);
    bool isMobileEqualN(mob_adress mob, double f);
    double *getFloatRef(string, int mode=0);
    mob_adress *getMobileRef(string, int mode=0);
    double *getFloatWriteRefCheck(string, int mode=0);
    mob_adress *getMobileWriteRefCheck(string, int mode=0);
    void setAttributes(atr_nfo& atr, double a);
}

// ------- regles du parseur d'AST (i.e. execution des noeuds)

start
    : ( expr_ )*
    ;

expr_
{}
  : (e:expr)
  { 
    MyASTNode *q = static_cast<MyASTNode*>((ANTLR_USE_NAMESPACE(antlr)AST*)e);
    cur_line = q->getLine();
  }
  ;


// ----- execution des commandes ------------------
expr
{
    double    a, b, cond;
    atr_nfo  atr;
    mob_adress mob;
}
    : #("if" cond=F:fexpr)
        {
            if( cond != 0.0 )
	    {
                if( F->getNextSibling() )
		    if( F->getNextSibling()->getFirstChild() )
			start(F->getNextSibling()->getFirstChild());
	    }
	    else
	    {
                if( F->getNextSibling() )
		{
		    if( F->getNextSibling()->getNextSibling() )
		    {
			start( F->getNextSibling()->getNextSibling() );
		    }
		}
	    }
        }
    | #("stop" mob=ref)
        {
            CMobile * mref =
	      getMobilePtr(mob, NOT_PSEUDOMOBILES|NOT_ME_IF_START);
	    if (!mref)
	      warning(MSG_null_ref);
	    else
	      mref->execStop();
        }
    | #("start" mob=ref)
        {
            CMobile * mref =
	      getMobilePtr(mob, NOT_PSEUDOMOBILES|NOT_ME_IF_START);
	    if (!mref)
	      warning(MSG_null_ref);
	    else
	      mref->execStart();
        }
    | #("restart" mob=ref)
        {
            CMobile * mref =
	      getMobilePtr(mob, NOT_PSEUDOMOBILES|NOT_ME_IF_START);
	    if (!mref)
	      warning(MSG_null_ref);
	    else 
                mref->execRestart();
        }
    | #("like" mob = ref)
        {
            CMobile * mref =
	      getMobilePtr(mob, NOT_PSEUDOMOBILES | NOT_ME);
	    if (!mref)
	      warning(MSG_null_ref);
	    else 
	      mref->startParserForField(this, field);
        }
    | #("move_to" mob=ref a=fexpr b=fexpr)
        {
            CMobile * mref = NULL;
            mref = getMobilePtr(mob,MOVEABLE);
	    if (!mref) warning(MSG_null_ref);
	    else mref->execMoveTo(a, b);
        }
    | (#("echo" ID))=> #("echo" id3:ID)
        { cerr << id3->getText() <<"\n"; }

// ----- affectation des variables globales ------------------

   | (#(COLON "trace" fexpr)) => #(COLON "trace" a=fexpr)
        {
	  WORLD.getCamera()->setVisible(a);
	}

// ----- affectation des variables temporaires ------------------
// NB: une variable peut etre une ref ou une fexp, donc on
// doit tester explicitement les cas.
// NB2: rappel: les INT sont consideres comme des REFs !

   | (#(COLON ID ID))=> #(COLON id_1:ID id_2:ID)
       {
	 double *fref1,*fref2;
	 mob_adress *mref1, *mref2;
	 fref2 = getFloatRef(id_2->getText());
	 if (fref2) {
	   fref1 = getFloatWriteRefCheck(id_1->getText(), CREATE_ON_FAIL);
	   if(fref1) *fref1 = *fref2;
	 }
	 else {
	   mref2 = getMobileRef(id_2->getText(), MUST_EXIST);
	   mref1 = getMobileWriteRefCheck(id_1->getText(), CREATE_ON_FAIL);
            if(mref1) *mref1 = *mref2;
        }
       }
    | (#(COLON ID fexpr))=> #(COLON id2:ID a=fexpr)
        {
            double * fref = NULL;
            fref = getFloatWriteRefCheck(id2->getText(), CREATE_ON_FAIL);
            if(fref) *fref = a;
        }
    | (#(COLON ID ref))=> #(COLON id:ID mob=ref)
        {
            mob_adress * mref = NULL;
            mref = getMobileWriteRefCheck(id->getText(), CREATE_ON_FAIL);
            if(mref) *mref = mob;
        }
  
// ----- affectation des attributs ------------------
    | (#(COLON attribref ID))=> #(COLON atr=attribref id_3:ID)
        {
	  double *fref = getFloatRef(id_3->getText());
	  if (fref)
	  setAttributes(atr, *fref);
	  else // cas "attrb:ref"
	  {
	    if (atr.name == "mem") {
	      mob_adress *mref = getMobileRef(id_3->getText(),MUST_EXIST);
	      // faute de savoir faire mem : ref, on traite au moins
	      // le cas ref=local (cast en int).
	      double a = conv_ref_to_number(*mref);
	      atr.m->setVar(atr.variable, a);

	    }
	    else throw CException(MSG_undef_f);
	  }
	}
    | (#(COLON attribref fexpr))=> #(COLON atr=attribref a=fexpr)
        {
	  setAttributes(atr, a);
	}
    | (#(COLON attribref ref))=> #(COLON atr=attribref mob=ref)
        {
	  if (atr.name == "mem") {
	    // cerr << "mem:ref \n";
	    double a = conv_ref_to_number(mob);
	    // faute de savoir faire mem : ref, on traite au moins
	    // le cas ref=local (cast en int).
	    atr.m->setVar(atr.variable, a);
	  }
	}
    ;

ref returns [mob_adress mob]
{
    double i,j,dy,ang;
}
    : #(REF i=fexpr j=fexpr dy=fexpr ang=fexpr)
        {
            mob.mobile = (int)i;
            mob.client = (int)j;
	    mob.dx = j;
	    mob.dy = dy;
	    mob.angle = ang;
        }
    | id:ID
        {
	  // cerr << "[DEBUG] ID=ref\n";
            mob_adress *mref = getMobileRef(id->getText());
            if( mref )
                mob = *mref;
            else
	    {
	        double *floatref = getFloatRef(id->getText(), MUST_EXIST);
		if( floatref != NULL )
		{
		  mob.mobile = ((int)(*floatref)); // cast (ref)fexp
		  mob.client = -1;
		}
            	else
            	{
		  mob.mobile = -100;
		  mob.client = -100;
           	}
	    }
        }
    ;

// an attrib as a left value
attribref returns [atr_nfo nfo]
{
    mob_adress mob;
}
    : #("x" mob=ref)  { set_atr_nfo(nfo, mob.mobile, mob.client, "x",  0); }
    | #("y" mob=ref)  { set_atr_nfo(nfo, mob.mobile, mob.client, "y",  0); }
    | #("dx" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "dx", 0); }
    | #("dy" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "dy", 0); }
    | #("angle" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "angle", 0); }
    | #("icon"  mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "icon", 0); }
    | #("color" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "color", 0); }
    | #("red"   mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "red", 0); }
    | #("green" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "green", 0); }
    | #("blue"  mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "blue", 0); }
    | #("gray"  mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "gray", 0); }
    | #("shockx" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "shockx", 0); }
    | #("shocky" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "shocky", 0); }
    | #("width"  mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "width", 0); }
    | #("height" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "height", 0); }
    | #("zoom"   mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "zoom", 0); }
    | #("visible" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "visible", 0); }
    | #("contact" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "contact", 0); }
    | #(i:"mem" mob=ref) { set_atr_nfo(nfo, mob.mobile, mob.client, "mem", atof(i->getText().c_str())); }
    ;

// ----- fonctions a evaluer ------------------
// "right" values
fexpr returns [double r]
{
    double a, b;
    mob_adress mob, mob2;
}
// --- operations -----
    : (#(PLUS fexpr fexpr)) => #(PLUS a=fexpr b=fexpr) { r = a + b; }
    | #(PLUS a=fexpr ) { r = a; }
    | (#(MINUS fexpr fexpr))=> #(MINUS a=fexpr b=fexpr) { r =  a - b; }
    | #(MINUS a=fexpr)        { r = -a;    }
    | #(STAR a=fexpr b=fexpr) { r = a * b; }
    | #(DIV a=fexpr b=fexpr)  {
         if( b != 0.0 ) r = a / b;
	 else { r = 0; warning(MSG_div0); }
       }
    | f:FLOAT { r = atof(f->getText().c_str()); }
    | i:INT   { r = atof(i->getText().c_str()); }
    | #("sqr" a=fexpr) { r = a * a;   }
    | #("sqrt" a=fexpr){ r = sqrt(fabs(a)); }
    | #("abs" a=fexpr) { r = fabs(a); }
    | #("sin" a=fexpr) { r = sin(a);  }
    | #("cos" a=fexpr) { r = cos(a);  }
    | #("tan" a=fexpr) { r = tan(a);  }
    | #("acs" a=fexpr) { r = acos(a); }
    | #("asn" a=fexpr) { r = asin(a); }
    | #("exp" a=fexpr) { r = exp(a);  }
    | #("log" a=fexpr) { r = log(fabs(a));  }
    | ( "clic" | "click" )   { r = WORLD.getClic();   }
    | "unclick"              { r = WORLD.getUnClic(); }
    | "clicked"              { r = WORLD.getClicked();}
    | "key_left"  { r = WORLD.isKeyPressed(KEY_LEFT); }
    | "key_right" { r = WORLD.isKeyPressed(KEY_RIGHT);}
    | "key_up"    { r = WORLD.isKeyPressed(KEY_UP);   }
    | "key_down"  { r = WORLD.isKeyPressed(KEY_DOWN); }
    | "key_space" { r = WORLD.isKeyPressed(KEY_SPACE);}
    | "rnd"  { r = drand48(); }
    | "srnd" { r = 2.0 * drand48() - 1.0; }
    | #("ang" a=fexpr b=fexpr) { r = atan2(b,a); }
    | #("min"  a=fexpr b=fexpr) { r = (a < b) ? a : b; }
    | #("max"  a=fexpr b=fexpr) { r = (a > b) ? a : b; }
    | #("int"  a=fexpr) { r = floor(a);   }
    | #("frac" a=fexpr) { r = a-floor(a); }
    | #("sgn"  a=fexpr) { r = (a<0.) ? -1. : (a>0.)? 1. : 0.; }
    | #("norm" a=fexpr b=fexpr) { r = sqrt(a*a+b*b); }
    | #("mod"  a=fexpr b=fexpr)
        {
            int aa = (int)a;
            int bb = (int)b;
            if( bb == 0 )
                r = aa;
            else
                r = aa - floor((double)aa/(double)bb) * bb;
        }
    | #("dist" mob=ref mob2=ref)
        {
            CMobile * mref  = getMobilePtr(mob);
            CMobile * mref2 = getMobilePtr(mob2);
            if ( mref && mref2 )
                r = mref->getPos().dist(mref2->getPos());
            else
	      { r = 0.0; warning(MSG_null_ref); }
        }
// --- attributs et variables -----
    | "t"  { r = WORLD.getTime(); } 
    | "dt" { r = WORLD.getTimeInterval(); } 
    |  #("x"  mob=ref) { r = get_atr_nfo(ATTR_X,  mob, 0); }
    |  #("y"  mob=ref) { r = get_atr_nfo(ATTR_Y,  mob, 0); }
    |  #("dx" mob=ref) { r = get_atr_nfo(ATTR_DX, mob, 0); }
    |  #("dy" mob=ref) { r = get_atr_nfo(ATTR_DY, mob, 0); }
    |  #("angle" mob=ref) { r = get_atr_nfo(ATTR_ANGLE, mob, 0); }
    |  #("icon"  mob=ref) { r = get_atr_nfo(ATTR_ICONE, mob, 0); }
    |  #("color" mob=ref) { r = get_atr_nfo(ATTR_COULEUR, mob, 0); }
    |  #("red"   mob=ref) { r = get_atr_nfo(ATTR_ROUGE, mob, 0); }
    |  #("green" mob=ref) { r = get_atr_nfo(ATTR_VERT, mob, 0);  }
    |  #("blue"  mob=ref) { r = get_atr_nfo(ATTR_BLEU, mob, 0);  }
    |  #("gray"  mob=ref) { r = get_atr_nfo(ATTR_GRIS, mob, 0);  }
    |  #("shockx" mob=ref) { r = get_atr_nfo(ATTR_CHOCX, mob, 0);   }
    |  #("shocky" mob=ref) { r = get_atr_nfo(ATTR_CHOCY, mob, 0);   }
    |  #("width"  mob=ref) { r = get_atr_nfo(ATTR_LARGEUR, mob, 0); }
    |  #("height" mob=ref) { r = get_atr_nfo(ATTR_HAUTEUR, mob, 0); }
    |  #("zoom"   mob=ref) { r = get_atr_nfo(ATTR_HAUTEUR, mob, 0); }
    |  #("visible" mob=ref) { r = get_atr_nfo(ATTR_VISIBLE, mob, 0); }
    |  #("contact" mob=ref) { r = get_atr_nfo(ATTR_CONTACT, mob, 0); }
    | #(va:"mem" mob=ref)   { r = get_atr_nfo(ATTR_VAR, mob, atoi(va->getText().c_str())); }

// --- tests d'egalite -----
    | (#(EQUAL ID ID))=> #(EQUAL id1:ID id2:ID)
    {
      // std::cerr << "[DEBUG] id EQUAL id ?\n";
      mob_adress *mref1 = getMobileRef(id1->getText());
      mob_adress *mref2 = getMobileRef(id2->getText());
      if (mref1 && mref2)
        r = (areMobilesEqual(*mref1, *mref2)) ? 1.0 : 0.0;
      else
      if ((!mref1)&&(!mref2)) {
	double *floatref1 = getFloatRef(id1->getText(),MUST_EXIST);
	double *floatref2 = getFloatRef(id2->getText(),MUST_EXIST);
	r = (*floatref1 == *floatref2) ? 1.0 : 0.0;
      }
      else
      if (mref1) 
        r = (isMobileEqualN(*mref1, *getFloatRef(id2->getText(),MUST_EXIST))) ? 1.0 : 0.0;
      else
        r = (isMobileEqualN(*mref2, *getFloatRef(id1->getText(),MUST_EXIST))) ? 1.0 : 0.0;
    }
    | (#(EQUAL ID ref))=> #(EQUAL id3:ID mob2=ref)
    {
      mob_adress *mref = getMobileRef(id3->getText());
      if (mref)
        r = (areMobilesEqual(*mref, mob2)) ? 1.0 : 0.0;
      else {
	double* floatref = getFloatRef(id3->getText(),MUST_EXIST);
	// printf("id=ref %d\n",*floatref==(int)*floatref);
	r = (*floatref == conv_ref_to_number(mob2)) ? 1.0 : 0.0;
      }
    }
    | (#(EQUAL ref ID))=> #(EQUAL mob2=ref id4:ID)
    {
      // std::cerr << "[DEBUG] ref EQUAL id ?\n";
      mob_adress *mref = getMobileRef(id4->getText());
      if (mref) {
        r = (areMobilesEqual(*mref, mob2)) ? 1.0 : 0.0;
      }
      else {
	double* floatref = getFloatRef(id4->getText(),MUST_EXIST);
	r = ((*floatref) == conv_ref_to_number(mob2)) ? 1.0 : 0.0;
      }
    }
    | (#(EQUAL fexpr fexpr))=> #(EQUAL a=fexpr b=fexpr)
    {
      // std::cerr << "[DEBUG] fexpr EQUAL fexpr ?\n";
      if( a == b ) r = 1.0; else r = 0.0;
    }
     | (#(EQUAL ref ref))=> #(EQUAL mob=ref mob2=ref)
      {
	// std::cerr << "[DEBUG] ref EQUAL ref ?\n";
	if (areMobilesEqual(mob,mob2))  r = 1.0; else r = 0.0;
	// if( getMobilePtr(mob) == getMobilePtr(mob2) ) r = 1.0; else r = 0.0;
      }
     | (#(EQUAL ref fexpr))=> #(EQUAL mob=ref a=fexpr)
      {
	// std::cerr << "[DEBUG] ref EQUAL fexpr ?\n";
	if (isMobileEqualN(mob,a))   r = 1.0; else r = 0.0;
      }
     | (#(EQUAL fexpr ref))=> #(EQUAL a=fexpr mob=ref)
      {
	// std::cerr << "[DEBUG] fexpr EQUAL ref ?\n";
	if (isMobileEqualN(mob,a))   r = 1.0; else r = 0.0;
      }

// --- tests d'inegalite -----
    | (#(NEQ ID ID))=> #(NEQ id_1:ID id_2:ID)
    {
      mob_adress *mref1 = getMobileRef(id_1->getText());
      mob_adress *mref2 = getMobileRef(id_2->getText());
      if (mref1 && mref2)
        r = (!areMobilesEqual(*mref1, *mref2)) ? 1.0 : 0.0;
      else
      if ((!mref1)&&(!mref2)) {
	double *floatref1 = getFloatRef(id_1->getText(),MUST_EXIST);
	double *floatref2 = getFloatRef(id_2->getText(),MUST_EXIST);
	r = (*floatref1 != *floatref2) ? 1.0 : 0.0;
      }
      else
      if (mref1) 
        r = (!isMobileEqualN(*mref1, *getFloatRef(id_2->getText(),MUST_EXIST))) ? 1.0 : 0.0;
      else
        r = (!isMobileEqualN(*mref2, *getFloatRef(id_1->getText(),MUST_EXIST))) ? 1.0 : 0.0;
    }
    | (#(NEQ ID ref))=> #(NEQ id5:ID mob2=ref)
    {
      mob_adress *mref = getMobileRef(id5->getText());
      if (mref)
        r = (!areMobilesEqual(*mref, mob2)) ? 1.0 : 0.0;
      else {
	double* floatref = getFloatRef(id5->getText(),MUST_EXIST);
	r = (*floatref != conv_ref_to_number(mob2)) ? 1.0 : 0.0;
      }
    }
    | (#(NEQ ref ID))=> #(NEQ mob2=ref id6:ID)
    {
      mob_adress *mref = getMobileRef(id6->getText());
      if (mref)
        r = (!areMobilesEqual(*mref, mob2)) ? 1.0 : 0.0;
      else {
	double* floatref = getFloatRef(id6->getText(),MUST_EXIST);
	r = (*floatref != conv_ref_to_number(mob2)) ? 1.0 : 0.0;
      }
    }
    | (#(NEQ fexpr fexpr))=> #(NEQ a=fexpr b=fexpr)
      { if( a != b ) r = 1.0; else r = 0.0; }
    | (#(NEQ ref ref))=> #(NEQ mob=ref mob2=ref)
      {
	  if (!areMobilesEqual(mob,mob2))  r = 1.0; else r = 0.0;
          // if( getMobilePtr(mob) != getMobilePtr(mob2) ) r = 1.0; else r = 0.0;
      }

// --- autres tests de comparaison -----
    | #(GREATER a=fexpr b=fexpr) { if( a >  b ) r = 1.0; else r = 0.0; }
    | #(GEQ     a=fexpr b=fexpr) { if( a >= b ) r = 1.0; else r = 0.0; }
    | #(LESS    a=fexpr b=fexpr) { if( a <  b ) r = 1.0; else r = 0.0; }
    | #(LEQ     a=fexpr b=fexpr) { if( a <= b ) r = 1.0; else r = 0.0; }
    | #("and"   a=fexpr b=fexpr) { if( ( a != 0.0 ) && ( b != 0.0 ) ) r = 1.0; else r = 0.0; }
    | #("or"    a=fexpr b=fexpr) { if( ( a != 0.0 ) || ( b != 0.0 ) ) r = 1.0; else r = 0.0; }

// --- variable -----
    |  id:ID
        {
            double * fref = getFloatRef(id->getText(), MUST_EXIST);
	    // cerr << "[DEBUG] ID=fexp\n";
            if( fref )
                r = *fref;
            else
	      { r = 0.0; warning(MSG_undef_f); }
        }
    ;










/*  **************************************************************
    LE PARSER : construct the AST tree (Abstract Syntax Tree)
*/




class CalcParser extends Parser;
options {
	buildAST = true;
	defaultErrorHandler=false;
}

start
    : ( (expr)? ( NEWLINE! | SEMI! ) )*
    ;

isolated_start
    : a:start
	{ #isolated_start = #([REF, "BIDON"], a); }
    ;

expr
    : (ID COLON^ fullref)=> ID COLON^ fullref
    | (lvalue COLON^ fexpr) => lvalue COLON^ fexpr
    | lvalue COLON^ ref
    | "if"^ fexpr "then"! isolated_start ( "else"! start )? "endif"!
    | "stop"^  ref
    | "start"^  ref
    | "restart"^ ref
    | "like"^  localref
    | "move_to"^ LPAREN! ref COMMA! fexpr COMMA! fexpr RPAREN!
    | ("trace" COLON^ fexpr) => "trace" COLON^ fexpr
    | "echo"^  ID
    ;


lvalue
//    :! va2:VAR
//        { #lvalue = #(va2, #([REF, "REF"], [INT,"-1"], [INT,"-1"], [INT,"-1"], [INT,"-1"])); }
    :! (attribute localref)=> (a:attribute l:localref)
        { #lvalue = #(a, l); }
    |! b:attribute
        { #lvalue = #(b, #([REF, "REF"], [INT,"-1"], [INT,"-1"], [INT,"-1"], [INT,"-1"])); }
    | ID
    ;

attribute
    : "dx" | "dy" | "x" | "y" | "angle" | "icon" | "width" | "height" | "zoom" | "color" | "red" | "green" | "blue" | "gray" | "shockx" | "shocky" | "visible" | "contact"
    |! "mem" i:INT
        { #attribute = #([LITERAL_mem, i->getText()]); }
    ;

localref
    : i1:INT
        { #localref = #([REF, "REF"], i1, [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "me"
        { #localref = #([REF, "REF"], [INT,"-1"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "prev"
        { #localref = #([REF, "REF"], [INT,"-4"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "next"
        { #localref = #([REF, "REF"], [INT,"-5"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "EMPTY"
        { #localref = #([REF, "REF"], [INT,"-13"], [INT,"-1"], [INT,"-1"], [INT,"-1"]);  }
    |! "camera"
        { #localref = #([REF, "REF"], [INT,"-17"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "light_one"
        { #localref = #([REF, "REF"], [INT,"-18"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "light_two"
        { #localref = #([REF, "REF"], [INT,"-19"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "light_three"
        { #localref = #([REF, "REF"], [INT,"-20"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    | ID
        { //  std::cerr << "[DEBUG] localref is an ID :-)\n";
	}
    ;

fullref
    :! ( INT AT )=> ( i:INT  AT j:INT )
        { #fullref = #([REF, "REF"], i, j, [INT,"-1"], [INT,"-1"]); }
    |! "me"
        { #fullref = #([REF, "REF"], [INT,"-1"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "mouse"
        { #fullref = #([REF, "REF"], [INT,"-2"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! ("collider") //  | "c")
        { #fullref = #([REF, "REF"], [INT,"-3"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "prev"
        { #fullref = #([REF, "REF"], [INT,"-4"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "next"
        { #fullref = #([REF, "REF"], [INT,"-5"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "nearest"
        { #fullref = #([REF, "REF"], [INT,"-6"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "nearest_second"
        { #fullref = #([REF, "REF"], [INT,"-7"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "nearest_third"
        { #fullref = #([REF, "REF"], [INT,"-8"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "nearest_left"
        { #fullref = #([REF, "REF"], [INT,"-9"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "nearest_right"
        { #fullref = #([REF, "REF"], [INT,"-10"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "nearest_up"
        { #fullref = #([REF, "REF"], [INT,"-11"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "nearest_down"
        { #fullref = #([REF, "REF"], [INT,"-12"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "EMPTY"
        { #fullref = #([REF, "REF"], [INT,"-13"], [INT,"-1"], [INT,"-1"], [INT,"-1"]);  }
    |! "nearest_in_dir" LPAREN dx:fexpr COMMA dy:fexpr COMMA ang:fexpr RPAREN
        { #fullref = #([REF, "REF"], [INT,"-14"], dx, dy, ang); }
    |! "nearest_in_dir_second" LPAREN dxA:fexpr COMMA dyA:fexpr COMMA angA:fexpr RPAREN
        { #fullref = #([REF, "REF"], [INT,"-15"], dxA, dyA, angA); }
    |! "nearest_in_dir_third" LPAREN dxB:fexpr COMMA dyB:fexpr COMMA angB:fexpr RPAREN
        { #fullref = #([REF, "REF"], [INT,"-16"], dxB, dyB, angB); }
    |! "camera"
        { #fullref = #([REF, "REF"], [INT,"-17"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "light_one"
        { #fullref = #([REF, "REF"], [INT,"-18"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "light_two"
        { #fullref = #([REF, "REF"], [INT,"-19"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
    |! "light_three"
        { #fullref = #([REF, "REF"], [INT,"-20"], [INT,"-1"], [INT,"-1"], [INT,"-1"]); }
//    | ID
    ;

ref
    :! (fullref)=> l:fullref
        { #ref = #l; }
//    |! ll:localref
    |! (localref) => ll:localref
        { #ref = #ll; }
    ;

fexpr : logicexpr ( ( "or"^ | "and"^ ) logicexpr )*
	 ;

logicexpr :  (ref (EQUAL | NEQ) ref) => ref ( EQUAL^ | NEQ^ ) ref
            { // std::cerr << "[DEBUG] Found a comparison between two refs\n";
	    }
      | addexpr ( ( EQUAL^ | GREATER^ | GEQ^ | LESS^ | LEQ^ | NEQ^ ) addexpr )*
	 ;

addexpr : mexpr ( ( PLUS^ | MINUS^ ) mexpr )*
    ;

mexpr : atom ( ( STAR^ | DIV^ | "mod"^ ) atom )*
    ;

atom
    : ( PLUS^ | MINUS^ ) atom
    | float_
    |! (attribute ref)=> a:attribute r:ref { #atom = #(a, r); }
    |! a1:attribute { #atom = #(a1, #([REF, "REF"], [INT,"-1"], [INT,"-1"], [INT,"-1"], [INT,"-1"])); }
    | LPAREN! fexpr RPAREN!
    |! ("PI"|"pi"|"Pi") { #atom =#([FLOAT, "3.14159"]); }
    |! "RED"   { #atom =#([FLOAT, RED   ]); }
    |! "GREEN" { #atom =#([FLOAT, GREEN ]); }
    |! "BLUE"  { #atom =#([FLOAT, BLUE  ]); }
    |! "YELLOW"{ #atom =#([FLOAT, YELLOW]); }
    |! "CYAN"  { #atom =#([FLOAT, CYAN  ]); }
    |! "PURPLE"{ #atom =#([FLOAT, PURPLE]); }
    |! "WHITE" { #atom =#([FLOAT, WHITE ]); }
    |! "BLACK" { #atom =#([FLOAT, BLACK ]); }
    |! "GRAY"  { #atom =#([FLOAT, GRAY  ]); }
    |! "ORANGE"{ #atom =#([FLOAT, ORANGE]); }
    | "sqrt"^ LPAREN! fexpr RPAREN!
    | "sqr"^  LPAREN! fexpr RPAREN!
    | "abs"^  LPAREN! fexpr RPAREN!
    | "sin"^  LPAREN! fexpr RPAREN!
    | "cos"^  LPAREN! fexpr RPAREN!
    | "tan"^  LPAREN! fexpr RPAREN!
    | "acs"^  LPAREN! fexpr RPAREN!
    | "asn"^  LPAREN! fexpr RPAREN!
    | "exp"^  LPAREN! fexpr RPAREN!
    | "log"^  LPAREN! fexpr RPAREN!
    | "rnd"^
    | "srnd"^
    | "clic"^
    | "click"^
    | "unclick"^
    | "clicked"^
    | "key_left"^
    | "key_right"^
    | "key_up"^
    | "key_down"^
    | "key_space"^
    | "ang"^  LPAREN! fexpr COMMA! fexpr RPAREN!
    | "min"^  LPAREN! fexpr COMMA! fexpr RPAREN!
    | "max"^  LPAREN! fexpr COMMA! fexpr RPAREN!
    | "int"^  LPAREN! fexpr RPAREN!
    | "frac"^ LPAREN! fexpr RPAREN!
    | "sgn"^  LPAREN! fexpr RPAREN!
    | "norm"^ LPAREN! fexpr COMMA! fexpr RPAREN!
    | "dist"^ LPAREN! ref   COMMA! ref   RPAREN!
    | "t"^
    | "dt"^
    | ID
    ;

float_
    : FLOAT
    | INT
    ;













/*  **************************************************************
    LE LEXEUR
    definition du nom des attributs, des symboles, des synonymes,
    et de la traduction des mots dans une autre langue.
*/





class CalcLexer extends Lexer;

options {
    testLiterals = false;
    charVocabulary = '\3'..'\377';
    defaultErrorHandler=false;
    k=17; // nombre de caracteres a tester pour les mots
}

COMMENT options { paraphrase = "un commentaire"; }
  :  "//" (~('\n'|'\r'))*
    { _ttype = ANTLR_USE_NAMESPACE(antlr)Token::SKIP; }
  ;

//protected
//STRING
//       : (~('\n'|'\r'))*
//       ;

protected
DIGITS options { paraphrase = "des chiffres"; }
	:	( '0'..'9' )+
	;

INT options { paraphrase = "un nombre"; }
: ('.')=> '.' DIGITS
         { $setType(FLOAT); }
| (DIGITS '.')=> DIGITS '.' ( '0'..'9' )*
        { $setType(FLOAT); }
| DIGITS
;

NEWLINE  options { paraphrase = "fin de la ligne";} : '\n'{newline();};
LPAREN   options { paraphrase = "'('";} :   '(';
RPAREN   options { paraphrase = "')'";} :   ')';
STAR     options { paraphrase = "'*'";} :   '*';
DIV      options { paraphrase = "'/'";} :   '/';
PLUS     options { paraphrase = "'+'";} :   '+';
MINUS    options { paraphrase = "'-'";} :   '-';
SEMI     options { paraphrase = "';'";} :   ';';
COLON    options { paraphrase = "':'";} :   ':';
COMMA    options { paraphrase = "','";} :   ',';
AT       options { paraphrase = "'@'";} :   '@';
EQUAL    options { paraphrase = "'='";} :   '=';
GREATER  options { paraphrase = "'>'";} :   '>';
LESS     options { paraphrase = "'<'";} :   '<';
GEQ      options { paraphrase = "'>='";} :  ">=";
LEQ      options { paraphrase = "'<='";} :  "<=";
NEQ      options { paraphrase = "'<>'";} :  "<>";

protected
LETTER options { paraphrase = "des lettres"; }
    : ( 'a'..'z' | 'A'..'Z' | '_' )
    ;

protected
ATTRIBUTES options { paraphrase = "un attribut"; }
: ( "dx" | "dy" | 'x' | 'y' | "angle" | "icon" | "mem" | "width" | "height" | "zoom" | "color" | "red" | "green" | "blue" | "gray" | "shockx" | "shocky" | "visible" | "contact" );

protected
SYNONYMS: "grey" | | "GREY" | "this" | "NULL" | "atan2"
    | "light" | "light1" | "light2" | "light3"
    | "nearest1" | "nearest2" | "nearest3"
    | "nearest_indir" | "nearest_indir1" | "nearest_indir2" | "nearest_indir3";

protected
FRENCH_VOCAB: "comme" | "si" | "alors" | "sinon" | "finsi" | "met_en"
    | "icone" | "hauteur" | "largeur" | "couleur" | "chocx" | "chocy"
    | "rouge" | "vert" | "bleu" | "gris"
    | "ROUGE" | "VERT" | "BLEU" | "JAUNE" | "MAGENTA" | "VIOLET" | "CYAN" | "BLANC" | "NOIR" | "GRIS"
    | "racine" | "carre" | "norme" | "ent" | "et" | "ou"
    | "souris" | "mulot" | "moi" | "lui"
    | "lampe" | "lampe1" | "lampe2" | "lampe3" 
    | "plusproche" | "plusproche2" | "plusproche1" | "plusproche3"
    | "pp" | "pp1" | "pp2" | "pp3"
    | "plusproche_gauche" | "plusproche_droite" | "plusproche_haut" | "plusproche_bas"
    | "ppg" | "ppd" | "pph" | "ppb"
    | "prec" | "suiv" | "VIDE" | "declic" | "cliquee"
    | "touche_gauche" | "touche_droite" | "touche_haut" | "touche_bas" | "touche_espace"
    | "plusproche_vers" | "plusproche_vers1" | "plusproche_vers2" | "plusproche_vers3"
    | "ppv" | "ppv1" | "ppv2" | "ppv3"
    ;

ID  options { testLiterals = true; paraphrase = "un mot"; }
    : (ATTRIBUTES)=> (
	  "dx" { $setType(LITERAL_dx); }
	| "dy" { $setType(LITERAL_dy); }
	| 'x'  { $setType(LITERAL_x);  }
	| 'y'  { $setType(LITERAL_y);  }
	| "angle" { $setType(LITERAL_angle); }
	| "icon"  { $setType(LITERAL_icon);  }
	| "mem"   { $setType(LITERAL_mem);   }
	| "zoom"  { $setType(LITERAL_zoom);  }
	| "height"{ $setType(LITERAL_height);}
	| "width" { $setType(LITERAL_width); }
	| "color" { $setType(LITERAL_color); }
	| "red"   { $setType(LITERAL_red);   }
	| "green" { $setType(LITERAL_green); }
	| "blue"  { $setType(LITERAL_blue);  }
	| "gray"  { $setType(LITERAL_gray);  }
	| "shockx"  { $setType(LITERAL_shockx);  }
	| "shocky"  { $setType(LITERAL_shocky);  }
	| "visible" { $setType(LITERAL_visible); }
	| "contact" { $setType(LITERAL_contact); }
    )
    // SYNONYMS
    | (SYNONYMS)=>  (
	  "grey"  { $setType(LITERAL_gray);  }
	| "GREY"  { $setType(LITERAL_GRAY);  }
        | "this"  { $setType(LITERAL_me);    }
        | "NULL"  { $setType(LITERAL_EMPTY); }
	| "atan2" { $setType(LITERAL_ang);      }
	| "light"  { $setType(LITERAL_light_one);   }
	| "light1" { $setType(LITERAL_light_one);   }
	| "light2" { $setType(LITERAL_light_two);   }
	| "light3" { $setType(LITERAL_light_three); }
	| "nearest1" { $setType(LITERAL_nearest);        }
	| "nearest2" { $setType(LITERAL_nearest_second); }
	| "nearest3" { $setType(LITERAL_nearest_third);  }
	| "nearest_indir"  { $setType(LITERAL_nearest_in_dir);        }
	| "nearest_indir1" { $setType(LITERAL_nearest_in_dir);        }
	| "nearest_indir2" { $setType(LITERAL_nearest_in_dir_second); }
	| "nearest_indir3" { $setType(LITERAL_nearest_in_dir_third);  }
    )
    // TRADUCTIONS
    | (FRENCH_VOCAB)=>  (
	  "comme" { $setType(LITERAL_like); }
	| "si"    { $setType(LITERAL_if);   }
	| "alors" { $setType(LITERAL_then); }
	| "sinon" { $setType(LITERAL_else); }
	| "finsi" { $setType(LITERAL_endif);}
	| "met_en" { $setType(LITERAL_move_to); }
	| "icone"   { $setType(LITERAL_icon);   }
	| "hauteur" { $setType(LITERAL_height); }
	| "largeur" { $setType(LITERAL_width);  }
	| "couleur" { $setType(LITERAL_color);  }
	| "chocx" { $setType(LITERAL_shockx); }
	| "chocy" { $setType(LITERAL_shocky); }
	| "rouge"{ $setType(LITERAL_red);  }
	| "vert" { $setType(LITERAL_green);}
	| "bleu" { $setType(LITERAL_blue); }
	| "gris" { $setType(LITERAL_gray); }
	| "ROUGE" { $setType(LITERAL_RED);   }
	| "VERT"  { $setType(LITERAL_GREEN); }
	| "BLEU"  { $setType(LITERAL_BLUE);  }
	| "JAUNE" { $setType(LITERAL_YELLOW);}
	| "MAGENTA"{ $setType(LITERAL_PURPLE);}
	| "VIOLET"{ $setType(LITERAL_PURPLE);}
	| "CYAN"  { $setType(LITERAL_CYAN);  }
	| "BLANC" { $setType(LITERAL_WHITE); }
	| "NOIR"  { $setType(LITERAL_BLACK); }
	| "GRIS"  { $setType(LITERAL_GRAY);  }
	| "racine"{ $setType(LITERAL_sqrt); }
	| "carre" { $setType(LITERAL_sqr);  }
	| "norme" { $setType(LITERAL_norm); }
	| "ent" { $setType(LITERAL_int); }
	| "et"  { $setType(LITERAL_and); }
	| "ou"  { $setType(LITERAL_or);  }
	| "souris"{ $setType(LITERAL_mouse); }
	| "mulot" { $setType(LITERAL_mouse); }
	| "lampe" { $setType(LITERAL_light_one);  }
	| "lampe1"{ $setType(LITERAL_light_one);  }
	| "lampe2"{ $setType(LITERAL_light_two);  }
	| "lampe3"{ $setType(LITERAL_light_three);}
	| "moi"   { $setType(LITERAL_me);         }
	| "lui"   { $setType(LITERAL_collider);   }
	| "plusproche"  { $setType(LITERAL_nearest);        }
	| "plusproche1" { $setType(LITERAL_nearest);        }
	| "plusproche2" { $setType(LITERAL_nearest_second); }
	| "plusproche3" { $setType(LITERAL_nearest_third);  }
	| "pp"  { $setType(LITERAL_nearest);        }
	| "pp1" { $setType(LITERAL_nearest);        }
	| "pp2" { $setType(LITERAL_nearest_second); }
	| "pp3" { $setType(LITERAL_nearest_third);  }
	| "plusproche_gauche" { $setType(LITERAL_nearest_left);  }
	| "plusproche_droite" { $setType(LITERAL_nearest_right); }
	| "plusproche_haut"   { $setType(LITERAL_nearest_up);    }
	| "plusproche_bas"    { $setType(LITERAL_nearest_down);  }
	| "ppg"   { $setType(LITERAL_nearest_left);  }
	| "ppd"   { $setType(LITERAL_nearest_right); }
	| "pph"   { $setType(LITERAL_nearest_up);    }
	| "ppb"   { $setType(LITERAL_nearest_down);  }
	| "prec" { $setType(LITERAL_prev); }
	| "suiv" { $setType(LITERAL_next); }
	| "VIDE" { $setType(LITERAL_EMPTY); }
	| "cliquee" { $setType(LITERAL_clicked); }
	| "declic"  { $setType(LITERAL_unclick); }
	| "touche_gauche" { $setType(LITERAL_key_left);   }
	| "touche_droite" { $setType(LITERAL_key_right);  }
	| "touche_haut"   { $setType(LITERAL_key_up);     }
	| "touche_bas"    { $setType(LITERAL_key_down);   }
	| "touche_espace" { $setType(LITERAL_key_space);  }
	| "plusproche_vers"  { $setType(LITERAL_nearest_in_dir);        }
	| "plusproche_vers1" { $setType(LITERAL_nearest_in_dir);        }
	| "plusproche_vers2" { $setType(LITERAL_nearest_in_dir_second); }
	| "plusproche_vers3" { $setType(LITERAL_nearest_in_dir_third);  }
	| "ppv"  { $setType(LITERAL_nearest_in_dir);        }
	| "ppv1" { $setType(LITERAL_nearest_in_dir);        }
	| "ppv2" { $setType(LITERAL_nearest_in_dir_second); }
	| "ppv3" { $setType(LITERAL_nearest_in_dir_third);  }
    )
    | LETTER ( LETTER | '0'..'9' )*
;


WS_ :   (   ' '
        |   '\t'
        |   '\r'
        )
        { _ttype = ANTLR_USE_NAMESPACE(antlr)Token::SKIP; } //ignore this token
;

